#pragma once

#include <raaz/hash/blake2/common.h>

const Word2s iv2s0 = 0x6a09e667UL;
const Word2s iv2s1 = 0xbb67ae85UL;
const Word2s iv2s2 = 0x3c6ef372UL;
const Word2s iv2s3 = 0xa54ff53aUL;
const Word2s iv2s4 = 0x510e527fUL;
const Word2s iv2s5 = 0x9b05688cUL;
const Word2s iv2s6 = 0x1f83d9abUL;
const Word2s iv2s7 = 0x5be0cd19UL;
