#pragma once

#include <raaz/hash/blake2/common.h>

const Word2b iv2b0 = 0x6a09e667f3bcc908ULL;
const Word2b iv2b1 = 0xbb67ae8584caa73bULL;
const Word2b iv2b2 = 0x3c6ef372fe94f82bULL;
const Word2b iv2b3 = 0xa54ff53a5f1d36f1ULL;
const Word2b iv2b4 = 0x510e527fade682d1ULL;
const Word2b iv2b5 = 0x9b05688c2b3e6c1fULL;
const Word2b iv2b6 = 0x1f83d9abfb41bd6bULL;
const Word2b iv2b7 = 0x5be0cd19137e2179ULL;
