-- | The AES block cipher.
module Raaz.Cipher.AES
       ( AES, KEY128, KEY192, KEY256, IV
       -- * Some AES cipher modes.
       , aes128cbc, aes192cbc, aes256cbc
       , aes128ctr
       ) where

import Raaz.Cipher.AES.Internal
import Raaz.Cipher.AES.Recommendation()

{-# ANN module "HLint: ignore Use import/export shortcut" #-}
